import yaml

def yaml_file(filename, load_all=False):
    with open(filename, mode='r') as f:
        if not load_all:
            yaml_data = yaml.load(f, Loader=yaml.FullLoader)
        else:
            yaml_data = list(yaml.load_all(f, Loader=yaml.FullLoader))  # convert generator to list before returning
    return yaml_data


def merge_data_objects(d1, d2, deduplicate_lists=True, merge_dict_in_lists=False):
    """A very simple function to merge 2 data objects into one recursively

    Args:
        d1 (dict or list): Original data object.
        d2 (dict or list): Data object to merge. Has higher priority in case of conflicting keys.
        deduplicate_lists (bool, optional): Remove duplicate elements from the list. Defaults to True.
        merge_dict_in_lists (bool, optional): Merge dictionaries from d2 into lists from d1. Defaults to False.

    Returns:
        dict or list: Merged data.
    """
    # a very simple function to merge 2 data objects into one recursively
    if isinstance(d1, dict) and isinstance(d2, dict):
        result = dict()
        all_keys = set(d1.keys()).union(d2.keys())
        for key in all_keys:
            if key not in d1.keys():
                result[key] = d2[key]
            elif key not in d2.keys():
                result[key] = d1[key]
            else:
                result[key] = merge_data_objects(d1[key], d2[key])
    elif isinstance(d1, list) and isinstance(d2, list):
        if deduplicate_lists:
            if merge_dict_in_lists:
                result = list()
                temp_dict = dict()
                for element in d1 + d2:
                    if isinstance(element, dict):
                        temp_dict = merge_data_objects(temp_dict, element)
                    else:
                        if element not in result:
                            result.append(element)
                result.append(temp_dict)
            else:
                result = list()
                for element in d1 + d2:
                    if element not in result:
                        result.append(element)
        else:
            result = d1 + d2
    else:  # 2nd object has higher priority if objects are different or not dict/list/str type
        result = d2

    return result



import jq
import json
import re
import sys
import glom
import runAM


class GlomDict(dict):

    def assign(self, glom_path, value):
        glom.assign(self, glom_path, value, missing=dict)


class PortConfigGenerator(runAM.db.JSONStore):

    def generatePortConfigData(self):
        """Generate low level data to parse port configuration templates.
        """
        # clean tables
        self.drop_table('server_tickets_expanded')  # this table contains server tickets with all profile data expanded
        self.drop_table('port_config_data')  # this table contains low level data required to parse port configuration templates
        # copy server_tickets table into server_tickets_expanded
        for doc_id, doc in self.table('server_tickets').items():
            self.insert_doc(doc, table_name='server_tickets_expanded')
        # find all profile tags combinations
        profile_list = self.jq(table_name='profile_tickets', query_expression='..|select(.tags?!=null)')
        # find (recursively) path to value in server_tickets_expanded table with profiles defined
        for path in self.jq_path(table_name='server_tickets_expanded', query_expression='..|select(.profiles?!=null)'):
            # get a value for the discovered path
            value_with_profile = self.get_val(path, table_name='server_tickets_expanded')
            for value_tags in value_with_profile['profiles']:
                profile_not_found = True  # if matching profile exists in profile_tickets table. If not raise and error.
                for profile in profile_list:
                    profile_tags = profile['tags']
                    if set(value_tags).issubset(set(profile_tags)):
                        profile_copy = profile.copy()  # to avoid modification of the original profile
                        del profile_copy['tags']  # delete tags from profile copy to avoid merging them with the ticket
                        # merge value with the profile data
                        value_with_profile = runAM.tools.merge_data_objects(value_with_profile, profile_copy)
                        profile_not_found = False
                if profile_not_found:
                    sys.exit(f'ERROR: Not able to find matching profile ticket for {value_tags} in profile_tickets table.')
            del value_with_profile['profiles']  # delete profiles as they are already expended
            self.update_path(path, value_with_profile, table_name='server_tickets_expanded')  # update table with the new value with merged profile data

        # build low level port configuration data
        # > check every server ticket
        for doc_id, server_ticket in self.table('server_tickets_expanded').items():
            server_ports_ticket = dict()

            # > if port-channel is configured and no group defined, generate LACP group number
            lacp_group_number = 0
            lacp_member_ports = list()
            for connection in server_ticket['connections']:
                if 'port_channel' in connection.keys():
                    # if group number is already defined manually in the ticket
                    if 'group_number' in connection['port_channel'].keys():
                        if not lacp_group_number:
                            lacp_group_number = connection['port_channel']['group_number']
                        elif lacp_group_number != connection['port_channel']['group_number']:
                            sys.exit(f'ERROR: statically defined LACP group numbers for {server_ticket["server_id"]} are not matching.')
                    else:
                        lacp_member_ports.append(connection['switch_port'])
            if lacp_member_ports:
                lacp_member_ports.sort()  # sort lacp members to find the lowest port
                lowest_lacp_member_port = lacp_member_ports[0]
                lowest_member_digits = ''.join(
                    re.findall(r'\d+', lowest_lacp_member_port)
                )
                if len(lowest_member_digits) == 1:
                    generated_group_number = int(f'100{lowest_member_digits}')
                elif len(lowest_member_digits) == 2:
                    generated_group_number = int(f'10{lowest_member_digits}')
                elif len(lowest_member_digits) == 3:
                    generated_group_number = int(f'1{lowest_member_digits}')
                else:
                    sys.exit('ERROR: Can not build port-channel name as memer interface name has more than 3 digits.')
                # check there is a conflict with a manually defined number defined for some connections by mistake
                if lacp_group_number:
                    if lacp_group_number != generated_group_number:
                        sys.exit(f'ERROR: LACP group number for {server_ticket["server_id"]} is defined manually, but not for all connections and not matching generated number.')
                lacp_group_number = generated_group_number  # if no error, use generated group number
            # build port-channel name
            port_channel_name = 'Port-Channel ' + str(lacp_group_number)

            # > check every server connection
            for connection in server_ticket['connections']:
                cd = GlomDict()  # cd - short for connection details

                # the switch name the node is connected to
                cd.assign('switch_name', connection['switch_name'])
                # define port description
                if 'description' in connection.keys():
                    description = connection['description']
                else:
                    description = server_ticket['server_id']  # default description is based on server ID
                cd.assign(f'interface.ethernet.{connection["switch_port"]}.description', description)
                # set interface speed
                if 'speed' in connection.keys():
                    speed = connection['speed']
                else:
                    speed = '25gfull'  # default interface speed is 25G
                cd.assign(f'interface.ethernet.{connection["switch_port"]}.speed', speed)
                # set FEC (forward error correction) for the interface if defined
                if 'fec' in connection.keys():
                    cd.assign(f'interface.ethernet.{connection["switch_port"]}.fec', connection['fec'])
                # set switchport configuration if defined
                if 'switchport' in connection.keys():
                    cd.assign(f'interface.ethernet.{connection["switch_port"]}.switchport', connection['switchport'])
                # build port-channel configuration
                if 'port_channel' in connection.keys():
                    cd.assign(f'interface.ethernet.{connection["switch_port"]}.channel.group', lacp_group_number)
                    cd.assign(f'interface.ethernet.{connection["switch_port"]}.channel.mode', connection['port_channel']['mode'])
                    cd.assign(f'interface.port_channel.{port_channel_name}.description', 'to_' + server_ticket["server_id"])
                    cd.assign(f'interface.port_channel.{port_channel_name}.mlag', lacp_group_number)
                    
                    if 'fallback' in connection['port_channel'].keys():
                        cd.assign(f'interface.port_channel.{port_channel_name}.fallback', connection['port_channel']['fallback'])

                    if 'lacp_rate' in connection['port_channel'].keys():
                        cd.assign(f'interface.ethernet.{connection["switch_port"]}.lacp.rate', connection['port_channel']['lacp_rate'])

                    if 'description' in connection['port_channel'].keys():
                        cd.assign(f'interface.port_channel.{port_channel_name}.description', connection['port_channel']['description'])

                    if 'switchport' in connection['port_channel'].keys():
                        cd.assign(f'interface.port_channel.{port_channel_name}.switchport', connection['port_channel']['switchport'])

                server_ports_ticket = runAM.tools.merge_data_objects(server_ports_ticket, cd)

            self.insert_doc(server_ports_ticket, table_name='port_config_data')
        


import argparse
import argcomplete
import runAM
import os
import sys
import inspect
import select
import json
from pprint import pprint

def add_arguments(a_parser, argument_list):
    """Add arguments to a parser

    Args:
        a_parser (argparse parser object): a parser to modify
        argument_list (list): list of agument specification dictionaries
    """

    for an_argument in argument_list:

        args = [
            an_argument['arg_name'],
            an_argument['arg_short_name'],
        ]

        kwargs = {'help': an_argument['help']}
        if 'action' in an_argument.keys():
            kwargs.update({'action': an_argument['action']})
        if 'default' in an_argument.keys():
            kwargs.update({'default': an_argument['default']})
        if 'choices' in an_argument.keys():
            kwargs.update({'choices': an_argument['choices']})
        if 'type' in an_argument.keys():
            kwargs.update({'type': eval(an_argument['type'])})  # eval to convert string to type
        if 'required' in an_argument.keys():
            kwargs.update({'required': an_argument['required']})

        a_parser.add_argument(*args, **kwargs)

def parse():
    # Build CLI for runAMcli from cli_spec.py
    # runAMcli is using argcomplete. To enable autocompletition, follow the docs here:
    # https://kislyuk.github.io/argcomplete/
    # TL;DR switch to bash and run: eval "$(register-python-argcomplete runAMcli)"
    parser = argparse.ArgumentParser(
        description=runAM.cli_spec['__main_parser']['help'], formatter_class=argparse.RawTextHelpFormatter)
    if 'add_argument' in runAM.cli_spec['__main_parser'].keys():
        add_arguments(parser, runAM.cli_spec['__main_parser']['add_argument'])
    subparsers = parser.add_subparsers(help="Select package", dest='subparser_name')

    for p_name, p_spec in sorted(runAM.cli_spec.items()):
        if p_name != '__main_parser':
            subparser = subparsers.add_parser(
                p_name, help=p_spec['help'],
                description=p_spec['help'],
                formatter_class=argparse.RawTextHelpFormatter
            )
            if 'add_argument' in p_spec.keys():
                add_arguments(subparser, p_spec['add_argument'])

    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    return vars(args)  # return args dictionary

def get_input(input_file=''):
    """Get input data from a file or stdin

    Args:
        input_file (str, optional): Filename to load input data. Must be json or YAML. Defaults to ''.

    Returns:
        [type]: [description]
    """

    # TODO: add CSV support

    # get stream from a file or stdin
    if input_file:
        in_stream = open(input_file, mode='r')
    else:
        if select.select([sys.stdin, ], [], [], 0.0)[0]:
            in_stream = sys.stdin

    # if input stream is present try to load data and add to all_params_dict
    try:
        in_string = "".join(in_stream)
    except Exception as _:
        pass  # ignore exception if in_stream not present or can not be loaded
    else:
        in_data = False

        try:  # try to load as JSON
            in_data = json.loads(in_string)
        except Exception as _:
            pass  # ignore if data is not a valid JSON

        if not in_data:
            try:  # try to load as YAML
                in_data = runAM.read.yaml_string(in_string, load_all=True)  # TODO: review if load_all is required?
            except Exception as _:
                pass  # ignore if data is not a valid YAML
        
        return in_data

def run_module(module_name_string, args_dict):
    """Run a module specified via CLI.

    Args:
        module_name_string (str): Module name to execute.
        args_dict (dict): CLI arguments dictionary.

    Returns:
        [any type returned by module]: Data returned by module. If any.
    """

    args = list()
    kwargs = dict()

    module_signature = inspect.signature(eval(module_name_string))
    for module_param_name in module_signature.parameters.keys():
        if module_signature.parameters[module_param_name].default is inspect._empty:
            # if default value is not defined, the parameter is positional and mandatory
            if module_param_name in args_dict.keys():
                args.append(args_dict[module_param_name])
            else:
                sys.exit('ERROR: Mandatory positional parameter {} is missing.'.format(module_param_name))
        else:
            if module_param_name in args_dict.keys():
                kwargs.update({
                    module_param_name: args_dict[module_param_name]
                })

    if module_name_string.startswith('runAM'):  # simple way to keep eval() safe
        return eval(module_name_string)(*args, **kwargs)

def interpreter():
    # process user input and execute a module
    args = runAM.cli.parse()

    # TODO: make module specific
    if 'input_file' in args.keys():
        if args['input_file']:
            in_data = runAM.cli.get_input(args['input_file'])
        else:
            in_data = runAM.cli.get_input()
    else:
        in_data = runAM.cli.get_input()

    args.update({
        'in_data': in_data
    })

    python_module_name = runAM.cli_spec[args['subparser_name']]['python_module']

    out_data = runAM.cli.run_module(python_module_name, args)
    if 'echo' in args.keys():
        if args['echo']:
            print(out_data)  # TODO: consider logging



import yaml

def yaml_string(a_string, load_all=False):
    if not load_all:
            yaml_data = yaml.load(a_string, Loader=yaml.FullLoader)
    else:
        yaml_data = list(yaml.load_all(a_string, Loader=yaml.FullLoader))  # convert generator to list before returning
    return yaml_data



import runAM
import sys

class ProfileTicketStore(runAM.db.JSONStore):

    def addSingleProfileTicket(self, ticket_data):
        """Add single ticket to profile_tickets table.

        Args:
            ticket_data (dict): Profile ticket data to add.
                                Verification:
                                - profile tags must be unique

        Returns:
            str: The inserted document ID.
        """
        inserted_doc_id = 0
        tag_list = ticket_data['tags']
        for _, doc in self.table('profile_tickets').items():
            if set(doc['tags']).issubset(tag_list):
                sys.exit(f'ERROR: A profile must have unique tags.')
        inserted_doc_id = self.insert_doc(ticket_data, table_name='profile_tickets')
        return inserted_doc_id

    def addProfileTicket(self, in_data):
        """Insert single or multiple profile tikets into the profile_tickets table.

        Args:
            in_data (dict or list): One or multiple tickets to insert.

        Returns:
            list: List of inserted document IDs.
        """
        doc_number_list = list()
        # for example, multi-doc yaml
        if isinstance(in_data, list):
            for ticket in in_data:
                doc_number = self.addSingleProfileTicket(ticket)
                if not doc_number:
                    sys.exit('ERROR: Failed to insert the ticket into the datababse.')
                else:
                    doc_number_list.append(doc_number)
        # for example, single-doc yaml
        elif isinstance(in_data, dict):
            doc_number = self.addSingleProfileTicket(in_data)
            if not doc_number:
                sys.exit('ERROR: Failed to insert the ticket into the datababse.')
            else:
                doc_number_list.append(doc_number)
        else:
            sys.exit('ERROR: Input data must be a dictionary or a list.')
        self.write()
        return doc_number_list

    def queryProfileTicket(self, tags):
        """Find profile tickets in profile_tickets table, that are matching specified tag list.

        Args:
            tags (str): List of tags to find matching profiles. Separated by ','.

        Returns:
            list: List of matching profiles in the format [{doc_id: doc}, ...]
        """
        tag_list = [tag.strip() for tag in tags.split(',')]
        profile_match_list = list()
        for doc_id, doc in self.table('profile_tickets').items():
            if set(tag_list).issubset(doc['tags']):
                profile_match_list.append({doc_id: doc})
        return profile_match_list

    def deleteProfileTicket(self, tags):
        """Delete profile tickets in profile_tickets table, that are matching specified tag list.

        Args:
            tags (str): List of tags to find matching profiles. Separated by ','.

        Returns:
            list: List of deleted document IDs.
        """
        tag_list = [tag.strip() for tag in tags.split(',')]
        deleted_doc_list = list()
        for doc_id, doc in self.table('profile_tickets').copy().items():  # iterate over table copy to avoid changing dict in the process
            if set(tag_list).issubset(doc['tags']):
                self.delete_doc(table_name='profile_tickets', doc_id=doc_id)
                deleted_doc_list.append(doc_id)
        self.write()
        return deleted_doc_list



# The cli_spec is used by runAM.cli to build build argparse arguments.

cli_spec = {
    '__main_parser': {
        'help': '=== runAM (run Arista Module) CLI help ===',
        'add_argument': [
            {
                'arg_name': '--echo',
                'arg_short_name': '-e',
                'action': 'store_true',
                'default': True,
                'help': 'Print child module output if available.',
            },
        ]
    },
    # server
    'server.add': {
        'python_module': 'runAM.ServerTicketStore().addServerTicket',
        'help': 'Add server ticket data.',
        'add_argument': [
            {
                'arg_name': '--input_file',
                'arg_short_name': '-inf',
                'help': 'Filename to load input data.',
            }
        ]
    },
    'server.delete': {
        'python_module': 'runAM.ServerTicketStore().deleteServerTicket',
        'help': 'Delete a server from the database.',
        'add_argument': [
            {
                'arg_name': '--server_id',
                'arg_short_name': '-sid',
                'help': 'Server to be deleted.',
            }
        ]
    },
    'server.query': {
        'python_module': 'runAM.ServerTicketStore().queryServerTicket',
        'help': 'Query server table by server ID.',
        'add_argument': [
            {
                'arg_name': '--server_id',
                'arg_short_name': '-sid',
                'help': 'Find a server with specified ID.',
            }
        ]
    },
    # profile
    'profile.add': {
        'python_module': 'runAM.ProfileTicketStore().addProfileTicket',
        'help': 'Add profile ticket data.',
        'add_argument': [
            {
                'arg_name': '--input_file',
                'arg_short_name': '-inf',
                'help': 'Filename to load input data.',
            }
        ]
    },
    'profile.delete': {
        'python_module': 'runAM.ProfileTicketStore().deleteProfileTicket',
        'help': 'Delete a profile from the database.',
        'add_argument': [
            {
                'arg_name': '--tags',
                'arg_short_name': '-t',
                'help': 'Tag list for a profile to be deleted. For example: --tags tag1, tag2, ...',
            }
        ]
    },
    'profile.query': {
        'python_module': 'runAM.ProfileTicketStore().queryProfileTicket',
        'help': 'Query profile table for specific tags.',
        'add_argument': [
            {
                'arg_name': '--tags',
                'arg_short_name': '-t',
                'help': 'Find a profile with specified tags. For example: --tags tag1, tag2, ...',
            }
        ]
    },
    # tools
    'tools.time_stamp': {
        'python_module': 'runAM.tools.time_stamp',
        'help': 'Print current time as Y-M-D H-M-S. In case you are lost in time and terminal has no time stamp enabled. =)',
    },
}



from time import time as time
from datetime import datetime as datetime

def time_stamp():
    """
    time_stamp function returns current system time as Y-M-D H-M-S string

    Args:
        no arguments required
    
    Returns:
        str: current system time as Y-M-D H-M-S string
    """
    time_not_formatted = time()
    time_formatted = datetime.fromtimestamp(time_not_formatted).strftime('%Y-%m-%d:%H:%M:%S.%f')
    return time_formatted



import runAM
import sys

class ServerTicketStore(runAM.generate.PortConfigGenerator):
    
    def addSingleServerTicket(self, ticket_data):
        """Add single ticket to server_tickets table.

        Args:
            ticket_data (dict): Server ticket data to add.
                                Verification:
                                - server_id must be unique
                                - switch_name/switch_port combination must not be in use

        Returns:
            str: The inserted document ID.
        """
        server_id = ticket_data['server_id']
        # check if server is already present
        for _, doc in self.table('server_tickets').items():
            if doc['server_id'] == server_id:
                sys.exit(f'ERROR: server {server_id} is already provisioned.')
        # check if port-leaf combination is already in use
        # TODO: Verify breakout vs parent port conflicts. For example: Ethernet1 and Ethernet1/1
        for ticket_connection in ticket_data['connections']:
            ticket_switch_name = ticket_connection['switch_name']
            ticket_switch_port = ticket_connection['switch_port']
            for _, doc in self.table('server_tickets').items():
                for existing_connection in doc['connections']:
                    if (ticket_switch_name == existing_connection['switch_name']) and (ticket_switch_port == existing_connection['switch_port']):
                        sys.exit(f'ERROR: Can not add {server_id}. Port {ticket_switch_port} on {ticket_switch_name} is already in use.')
        # if no verification failed, insert the ticket
        inserted_doc_id = self.insert_doc(ticket_data, table_name='server_tickets')
        return inserted_doc_id

    def addServerTicket(self, in_data):
        """Insert single or multiple server tikets into the server_tickets table.

        Args:
            in_data (dict or list): One or multiple tickets to insert.

        Returns:
            list: List of inserted document IDs.
        """
        doc_number_list = list()
        # for example, multi-doc yaml
        if isinstance(in_data, list):
            for ticket in in_data:
                doc_number = self.addSingleServerTicket(ticket)
                if not doc_number:
                    sys.exit('ERROR: Failed to insert the ticket into the datababse.')
                else:
                    doc_number_list.append(doc_number)
        # for example, single-doc yaml
        elif isinstance(in_data, dict):
            doc_number = self.addSingleServerTicket(in_data)
            if not doc_number:
                sys.exit('ERROR: Failed to insert the ticket into the datababse.')
            else:
                doc_number_list.append(doc_number)
        else:
            sys.exit('ERROR: Input data must be a dictionary or a list.')
        self.generatePortConfigData()  # generate low level port config data after inserting a new server
        self.write()
        return doc_number_list

    def queryServerTicket(self, server_id):
        """Find server ticket in server_tickets table, that is matching server_id.

        Args:
            server_id (str): Server ID to find in server_tickets table.

        Returns:
            list: List of matching servers in the format [{doc_id: doc}, ...]
        """
        server_match_list = list()
        for doc_id, doc in self.table('server_tickets').items():
            if doc['server_id'] == server_id:
                server_match_list.append({doc_id: doc})
        return server_match_list

    def deleteServerTicket(self, server_id):
        """Delete server ticket in server_tickets table, that is matching server_id.

        Args:
            server_id (str): Server ID to delete from server_tickets table.

        Returns:
            list: List of deleted document IDs.
        """
        deleted_doc_list = list()
        for doc_id, doc in self.table('server_tickets').copy().items():  # iterate over table copy to avoid changing dict in the process
            if doc['server_id'] == server_id:
                self.delete_doc(table_name='server_tickets', doc_id=doc_id)
                deleted_doc_list.append(doc_id)
        self.generatePortConfigData()  # generate low level port config data after inserting a new server
        self.write()
        return deleted_doc_list



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="runAM-evpn",
    version="0.1.1",
    description="runAM Python modules to build EVPN network",
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/arista-netdevops-community/runAM',
    author='Petr Ankudinov',
    license="BSD",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX :: Linux",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "argcomplete==1.12.1",
        "glom==20.11.0",
        "jq==1.1.1",
        "PyYAML==5.3.1",
    ],
    entry_points = {
        "console_scripts": ['runAM = runAM.cli:interpreter']
    },
)



#!/usr/bin/env python3
import jq
import json
import sys
import os
from runAM.tools.change_nested_value import change_nested_value

class JSONStore:

    def __init__(self, database_name = 'db', directory='database'):
        self.db_filename = os.path.join(directory, f'{database_name}.json')

        # try to load the file if exists
        try:
            with open(self.db_filename, 'r') as db_file:
                self.db = json.load(db_file)
        except Exception as _:
            # if nothing was opened, init as empty dictionary
            self.db = dict()

    def write(self):
        """Save data to the file on exit.

        Returns:
            True: return True after execution for assert tests
        """
        with open(self.db_filename, 'w') as db_file:
            json.dump(self.db, db_file, indent=4)
        return True

    def drop_table(self, table_name='_default'):
        """Init existing or non-existing table in the database with an empty dictionary.

        Args:
            table_name (str, optional): Name of the table to init. Defaults to '_default'.

        Returns:
            dict: Empty table - {}. 
        """
        self.db.update({table_name: dict()})  # init table with empty dict
        return self.table(table_name)

    def insert_doc(self, data, doc_id='', table_name='_default'):
        """Insert a document into the database.

        Args:
            data (any): Data to insert as a document in the table.
            doc_id (str, optional): Document ID to set manually. If not specified, new doc ID will be generated automatically.
            table_name (str, optional): Table name to insert the document. Defaults to '_default'.

        Returns:
            str: New document ID.
        """
        existing_doc_ids = [int(id) for id in self.table(table_name).keys() if str(id).isdecimal()]
        if doc_id:  # if doc_id was defined manually
            new_doc_id = doc_id
            if int(new_doc_id) in existing_doc_ids:
                sys.exit(f'ERROR: docID {doc_id} is already present in the table {table_name}')
        else:
            try:
                new_doc_id = [id for id in range(min(existing_doc_ids), max(existing_doc_ids)+2) if id not in existing_doc_ids][0]  # take lowest missing docID or next after max
            except ValueError:  # ValueError: min() arg is an empty sequence
                new_doc_id = 1  # if table is empty
        self.table(table_name).update({
            str(new_doc_id): data
        })
        return str(new_doc_id)

    def table(self, table_name):
        """Return table content.

        Args:
            table_name (str): The name of the table to be returned.

        Returns:
            dict: The table content as doc_id: doc_value dictionary.
        """
        if table_name not in self.db.keys():
            self.db.update({table_name: dict()})  # add empty table if not present
        return self.db[table_name]

    def get_table_names(self):
        """Get the list of all table names in the database.

        Returns:
            list: List of table names.
        """
        return list(self.db.keys())

    def jq(self, table_name='_default', query_expression=''):
        """Find a value in a table using jq: https://stedolan.github.io/jq/

        Args:
            table_name (str, optional): The name of the table to query. Defaults to '_default'.
            query_expression (str, optional): jq query expression. Defaults to ''.

        Returns:
            list: List of matched values.
        """
        res = jq.compile(query_expression).input(self.table(table_name)).all()
        return res

    def jq_path(self, table_name='_default', query_expression='', return_glom_spec=False):
        """Find a path to a matched value in a table using jq: https://stedolan.github.io/jq/

        Args:
            table_name (str, optional): The name of the table to query. Defaults to '_default'.
            query_expression (str, optional): jq query expression. Defaults to ''.
            return_glom_spec (bool, optional): Return glom (https://glom.readthedocs.io/) spec tring instead of a list. Defaults to False.

        Returns:
            list: List of pathes to every matched value. Every path is a list of keys/indexes starting at table root.
        """
        res = jq.compile('path(%s)' % query_expression).input(self.table(table_name)).all()
        if return_glom_spec:
            path_list = list()
            for a_path in res:
                glom_spec = ''
                for k in a_path:
                    glom_spec += f'.{k}'
                path_list.append(glom_spec.lstrip('.'))
        else:
            path_list = res
        return path_list

    def get_val(self, path_list, table_name='_default'):
        """Get the value that corresponds to the path calculated by jq.

        Args:
            path_list (list): List of pathes to every value to be returned. 
            table_name (str, optional): The name of the table to where matched values must be located. Defaults to '_default'.

        Returns:
            list: List of values corresponding to to specified pathes.
        """
        data = self.table(table_name)
        for element in path_list:
            data = data[element]
        return data

    def delete_doc(self, table_name='_default', doc_id=''):
        """Delete a document with the specified doc_id from the table.

        Args:
            table_name (str, optional): The name of the table to delete the document from. Defaults to '_default'.
            doc_id (str, optional): The document id to be deleted. Defaults to ''.

        Returns:
            list: List of document IDs to be deleted.
        """
        deleted_docs = list()
        if doc_id:
            del self.table(table_name)[doc_id]
            deleted_docs.append(doc_id)
        return deleted_docs

    def update_path(self, path, data=dict(), table_name='_default'):
        """Update the value corresponding to the specified path.

        Args:
            path (list): list of keys/indexes starting at table root
            data (any, optional): The new value to be set. Defaults to dict().
            table_name (str, optional): The name of the table where the value must be changed. Defaults to '_default'.

        Returns:
            dict: Updated table content.
        """
        # update value in a table based on specified path
        self.table(table_name).update(
            change_nested_value(path, self.table(table_name), data)
        )
        return self.table(table_name)



def change_nested_value(path, in_data, new_value=dict(), delete=False):
    """Change a value indexed by a key list.
    Example: change_nested_value(['k1', 0, 'k2'], some_data, some_value) is equal to
                some_data['k1'][0]['k2'] = some_value

    Args:
        path (list): List of keys or list indexes to find the path in nested data structure.
        in_data (dict|list): Input data to be changed.
        new_value (any): Value to be set.
        delete (bool, optional): Delete the element instead of changing the value. Defaults to False.

    Returns:
        dict|list: Updated data.
    """
    path_copy = path.copy()  # to avoid changing original list
    a_key = path_copy.pop(0)
    out_data = in_data.copy()
    if not len(path_copy):
        if delete:
            del out_data[a_key]
        else:
            out_data[a_key] = new_value
    else:
        out_data[a_key] = change_nested_value(path_copy, in_data[a_key], new_value, delete=delete)
    return out_data
